package cc.niushuai.projects.demo.aspectlock.common.anno;

import cc.niushuai.projects.demo.aspectlock.common.exception.LockException;
import cc.niushuai.projects.demo.aspectlock.common.util.AopLockUtil;
import cn.hutool.core.util.ReflectUtil;
import cn.hutool.core.util.StrUtil;
import lombok.extern.slf4j.Slf4j;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Pointcut;
import org.aspectj.lang.reflect.MethodSignature;
import org.springframework.stereotype.Component;

import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

/**
 * LockAspect 锁切面
 *
 * @author niushuai
 * @date 2021/12/22 10:17
 */
@Slf4j
@Aspect
@Component
public class LockAspect {

    private static final String POINT = ".";
    private static final String REGEX_POINT = "\\.";

    private static final Class[] BASE_TYPE_CLASS =
            new Class[]{
                    Byte.class,
                    Character.class,
                    Short.class, Integer.class, Long.class,
                    Float.class, Double.class,
                    String.class,
                    Boolean.class,
                    Date.class
            };

    @Pointcut(value = "@annotation(cc.niushuai.projects.demo.aspectlock.common.anno.AopLock)")
    public void lockPointcut() {
    }

    @Around(value = "lockPointcut()")
    public Object doAround(ProceedingJoinPoint joinPoint) throws Throwable {

        Lock lock = parseLock(joinPoint);
        try {
            // 加锁
            AopLockUtil.lock(lock);
            // 执行方法
            return joinPoint.proceed();
        } finally {
            // 解锁
            AopLockUtil.unlock(lock);
        }
    }

    /**
     * 封装出一个lock对象
     *
     * @param joinPoint 切入点
     * @author niushuai
     * @date: 2021/12/22 17:42
     * @return: {@link Lock} 封装过的lock对象
     */
    private Lock parseLock(ProceedingJoinPoint joinPoint) {
        Lock lock = new Lock();
        StringBuffer key = new StringBuffer(), value = new StringBuffer();

        // 实际参数值
        Object[] args = joinPoint.getArgs();
        MethodSignature method = (MethodSignature) joinPoint.getSignature();

        // 方法签名上的参数类型
        // Class[] parameterTypes = method.getParameterTypes();
        // 方法签名上的参数名称
        // String[] parameterNames = new String[]{};

        // 确定key的顺序 便于后续反射取值
        Map<String, Integer> paramOrderMap = transferParamOrderMap(method.getParameterNames());
        Map<Integer, Class> classOrderMap = transferClassOrderMap(method.getParameterTypes());

        // 方法上的注解 携带有具体的内容
        AopLock annotation = method.getMethod().getAnnotation(AopLock.class);
        String[] keyArr = annotation.key();
        String[] valueArr = annotation.value();
        // 根据是否启用反射取值来进行不同的处理
        log.debug("启用反射取值 - {}", annotation.enableReflect());
        if (annotation.enableReflect()) {
            // key
            reflectVal(key, args, paramOrderMap, classOrderMap, keyArr);
            // value
            reflectVal(value, args, paramOrderMap, classOrderMap, valueArr);
        } else {
            // 未启用反射 直接拼接key和value
            key.append(StrUtil.concat(true, annotation.prefix(), StrUtil.join(annotation.keySeparator(), keyArr)));
            value.append(StrUtil.join(annotation.valueSeparator(), valueArr));
        }
        log.debug("锁值 - key: {}, value: {}", key, value);

        lock.setKey(key.toString()).setValue(value.toString()).setPeriod(annotation.period());

        return lock;
    }

    private void reflectVal(StringBuffer key, Object[] args, Map<String, Integer> paramOrderMap, Map<Integer, Class> classOrderMap, String[] keyArr) {
        for (String tmpKey : keyArr) {
            // student.age
            // 要先判断 . 是否存在
            boolean baseType = false;
            if (tmpKey.contains(POINT)) {
                // 含有点
            } else if (judgeBaseType(tmpKey, paramOrderMap, classOrderMap)) {
                // 基础类型
                baseType = true;
            } else {
                throw new LockException("取值 - 在启用反射获取值时请按照paramName.fieldName的方式来做 [" + tmpKey + "]");
            }



            if (baseType) {
                // 基础类型不用反射  直接拼接即可
                key.append(args[paramOrderMap.get(tmpKey)]);
                continue;
            }

            // 获取paramName.fieldName的具体值
            String[] split = tmpKey.split(REGEX_POINT);
            String paramName = split[0];
            String fieldName = split[1];

            if (!paramOrderMap.containsKey(paramName)) {
                throw new LockException("取值 - 不存在形参名为[" + paramName + "]的参数");
            }

            try {
                // 非基础类型 通过反射取舍  像list map啊啥的 么想好怎么搞
                // 感觉可以通过类型判断来对于这些特定类型进行按照特定的方式进行取值
                // 最后才会使用到反射来取值  一般应用场景的话 形参对象居多 先不考虑吧
                Object fieldValue = ReflectUtil.getFieldValue(args[paramOrderMap.get(paramName)], fieldName);
                key.append(fieldValue);
            } catch (Exception e) {
                throw new LockException("取值 - 不存在形参名为[" + paramName + "], 且字段名称为[" + fieldName + "]的参数, 完整参数: [" + tmpKey + "]");
            }

        }
    }

    /**
     * 判断既定key是否为基础类型
     *
     * @param tmpKey        既定key
     * @param paramOrderMap 参数顺序
     * @param classOrderMap clazz顺序
     * @author niushuai
     * @date: 2021/12/23 10:03
     * @return: {@link boolean}
     */
    private boolean judgeBaseType(String tmpKey, Map<String, Integer> paramOrderMap, Map<Integer, Class> classOrderMap) {
        // 进入到此处时均为不带点的参数
        if (!paramOrderMap.containsKey(tmpKey)) {
            // 由于基础类型不需要. 因为tmpKey应于key保持一致 所以在paramOrderKey中含有则对 若不含有
            throw new LockException("取值 - [" + tmpKey + "]在形参中不存在");
        }

        Class clazz = classOrderMap.get(paramOrderMap.get(tmpKey));
        // 此处不可能为null 匹配既定的任一类型
        return Arrays.stream(BASE_TYPE_CLASS).anyMatch(item -> item.isAssignableFrom(clazz));
    }

    private Map<String, Integer> transferParamOrderMap(String[] parameterNames) {
        Map<String, Integer> map = new HashMap<>();

        for (int i = 0; i < parameterNames.length; i++) {
            map.put(parameterNames[i], i);
        }
        return map;
    }

    private Map<Integer, Class> transferClassOrderMap(Class[] clazz) {
        Map<Integer, Class> map = new HashMap<>();

        for (int i = 0; i < clazz.length; i++) {
            map.put(i, clazz[i]);
        }
        return map;
    }

}
