package cc.niushuai.projects.demo.aspectlock.common.util;

import cc.niushuai.projects.demo.aspectlock.common.anno.Lock;
import cc.niushuai.projects.demo.aspectlock.common.exception.LockException;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;

import java.util.HashMap;
import java.util.Map;

/**
 * 公共工具类
 *
 * @author niushuai
 * @date 2021/12/22 10:19
 */
public class AopLockUtil {

    /**
     * 锁柜
     */
    private static final Map<String, Lock> LOCK_MAP = new HashMap<>(16);

    private AopLockUtil() {
    }

    /**
     * 从锁柜中根据钥匙获取到相应的锁
     *
     * @param key 钥匙
     * @author niushuai
     * @date: 2021/12/22 11:24
     * @return: {@link Lock}
     */
    private static Lock getLock(String key) {
        return LOCK_MAP.get(key);
    }

    /**
     * 加锁
     *
     * @param lock 锁
     * @author niushuai
     * @date: 2021/12/22 11:25
     * @return: void
     */
    public static void lock(Lock lock) {

        // 加锁时 钥匙不能为空
        validate(lock.getKey());

        // 要求锁不存在
        if (isLocked(lock.getKey())) {
            throw new LockException("加锁失败 - 锁[" + lock.getKey() + "]已存在");
        }

        // 重置key过期时间
        lock.setExpireTime(getExpireTime(lock.getPeriod()));
        // 放入锁柜
        LOCK_MAP.put(lock.getKey(), lock);
    }


    /**
     * 解锁
     *
     * @param lock 锁
     * @author niushuai
     * @date: 2021/12/22 11:29
     * @return: void
     */
    public static void unlock(Lock lock) {
        // 要求锁存在
        if (ObjectUtil.isNull(getLock(lock.getKey()))) {
            throw new LockException("解锁失败 - 锁[" + lock.getKey() + "]不存在");
        }
        // 无需关心过期问题 直接移除锁柜
        LOCK_MAP.remove(lock.getKey());
    }

    /**
     * 是否可锁
     * 为空 可锁
     * 锁已过期 可锁
     *
     * @param key 钥匙
     * @author niushuai
     * @date: 2021/12/22 11:06
     * @return: {@link boolean}
     */
    public static boolean lockable(String key) {
        Lock lock = getLock(key);

        return ObjectUtil.isEmpty(lock) || lock.isExpire();
    }

    /**
     * 非空 且未过期 即为true
     * 非空 已过期 为false
     * 空  为false
     *
     * @param key 钥匙
     * @author niushuai
     * @date: 2021/12/22 11:04
     * @return: {@link boolean} 是否可锁
     */
    public static boolean isLocked(String key) {
        return !lockable(key);
    }


    private static void validate(String key) {
        if (StrUtil.isEmpty(key)) {
            throw new LockException("锁[" + key + "]不能为空");
        }
    }

    /**
     * 根据有效时间 获取基于当前时间的有效期
     *
     * @param period
     * @author niushuai
     * @date: 2021/12/22 10:40
     * @return: {@link long}
     */
    private static long getExpireTime(long period) {
        return System.currentTimeMillis() + period;
    }
}
